@@ -11,12 +11,16 @@ operations.  PropertyLoader also relies upon the dependency.py module
 to handle flush-time dependency sorting and processing.
 """
 
-from sqlalchemy import sql, schema, util, exceptions, sql_util, logging
+from sqlalchemy import sql, schema, util, exceptions, logging
+from sqlalchemy.sql import util as sql_util
 from sqlalchemy.orm import mapper, sync, strategies, attributes, dependency
 from sqlalchemy.orm import session as sessionlib
 from sqlalchemy.orm import util as mapperutil
-import sets, random
-from sqlalchemy.orm.interfaces import *
+import operator
+from sqlalchemy.orm.interfaces import StrategizedProperty, PropComparator
+from sqlalchemy.exceptions import ArgumentError
+
+__all__ = ['ColumnProperty', 'CompositeProperty', 'PropertyLoader', 'BackRef']
 
 class ColumnProperty(StrategizedProperty):
     """Describes an object attribute that corresponds to a table column."""
@@ -31,38 +35,101 @@ class ColumnProperty(StrategizedProperty):
         self.columns = list(columns)
         self.group = kwargs.pop('group', None)
         self.deferred = kwargs.pop('deferred', False)
-
+        self.comparator = ColumnProperty.ColumnComparator(self)
+        # sanity check
+        for col in columns:
+            if not hasattr(col, 'name'):
+                if hasattr(col, 'label'):
+                    raise ArgumentError('ColumnProperties must be named for the mapper to work with them.  Try .label() to fix this')
+                raise ArgumentError('%r is not a valid candidate for ColumnProperty' % col)
+        
     def create_strategy(self):
         if self.deferred:
             return strategies.DeferredColumnLoader(self)
         else:
             return strategies.ColumnLoader(self)
-
-    def getattr(self, object):
+    
+    def copy(self):
+        return ColumnProperty(deferred=self.deferred, group=self.group, *self.columns)
+        
+    def getattr(self, object, column):
         return getattr(object, self.key)
 
-    def setattr(self, object, value):
+    def setattr(self, object, value, column):
         setattr(object, self.key, value)
 
     def get_history(self, obj, passive=False):
         return sessionlib.attribute_manager.get_history(obj, self.key, passive=passive)
 
-    def merge(self, session, source, dest, _recursive):
+    def merge(self, session, source, dest, dont_load, _recursive):
         setattr(dest, self.key, getattr(source, self.key, None))
 
-    def compare(self, value):
-        return self.columns[0] == value
+    def get_col_value(self, column, value):
+        return value
 
+    class ColumnComparator(PropComparator):
+        def clause_element(self):
+            return self.prop.columns[0]
+            
+        def operate(self, op, *other):
+            return op(self.prop.columns[0], *other)
+
+        def reverse_operate(self, op, other):
+            col = self.prop.columns[0]
+            return op(col._bind_param(other), col)
+
+            
 ColumnProperty.logger = logging.class_logger(ColumnProperty)
 
-mapper.ColumnProperty = ColumnProperty
+
+class CompositeProperty(ColumnProperty):
+    """subclasses ColumnProperty to provide composite type support."""
+    
+    def __init__(self, class_, *columns, **kwargs):
+        super(CompositeProperty, self).__init__(*columns, **kwargs)
+        self.composite_class = class_
+        self.comparator = kwargs.pop('comparator', CompositeProperty.Comparator)(self)
+        
+    def copy(self):
+        return CompositeProperty(deferred=self.deferred, group=self.group, composite_class=self.composite_class, *self.columns)
+
+    def getattr(self, object, column):
+        obj = getattr(object, self.key)
+        return self.get_col_value(column, obj)
+
+    def setattr(self, object, value, column):
+        obj = getattr(object, self.key, None)
+        if obj is None:
+            obj = self.composite_class(*[None for c in self.columns])
+        for a, b in zip(self.columns, value.__composite_values__()):
+            if a is column:
+                setattr(obj, b, value)
+
+    def get_col_value(self, column, value):
+        for a, b in zip(self.columns, value.__composite_values__()):
+            if a is column:
+                return b
+
+    class Comparator(PropComparator):
+        def __eq__(self, other):
+            if other is None:
+                return sql.and_(*[a==None for a in self.prop.columns])
+            else:
+                return sql.and_(*[a==b for a, b in
+                                  zip(self.prop.columns,
+                                      other.__composite_values__())])
+
+        def __ne__(self, other):
+            return sql.or_(*[a!=b for a, b in
+                             zip(self.prop.columns,
+                                 other.__composite_values__())])
 
 class PropertyLoader(StrategizedProperty):
     """Describes an object property that holds a single item or list
     of items that correspond to a related database table.
     """
 
-    def __init__(self, argument, secondary, primaryjoin, secondaryjoin, entity_name=None, foreign_keys=None, foreignkey=None, uselist=None, private=False, association=None, order_by=False, attributeext=None, backref=None, is_backref=False, post_update=False, cascade=None, viewonly=False, lazy=True, collection_class=None, passive_deletes=False, remote_side=None, enable_typechecks=True):
+    def __init__(self, argument, secondary=None, primaryjoin=None, secondaryjoin=None, entity_name=None, foreign_keys=None, foreignkey=None, uselist=None, private=False, association=None, order_by=False, attributeext=None, backref=None, is_backref=False, post_update=False, cascade=None, viewonly=False, lazy=True, collection_class=None, passive_deletes=False, remote_side=None, enable_typechecks=True, join_depth=None, strategy_class=None):
         self.uselist = uselist
         self.argument = argument
         self.entity_name = entity_name
@@ -75,21 +142,32 @@ class PropertyLoader(StrategizedProperty):
         self.lazy = lazy
         self.foreign_keys = util.to_set(foreign_keys)
         self._legacy_foreignkey = util.to_set(foreignkey)
+        if foreignkey:
+            util.warn_deprecated('foreignkey option is deprecated; see docs for details')
         self.collection_class = collection_class
         self.passive_deletes = passive_deletes
         self.remote_side = util.to_set(remote_side)
         self.enable_typechecks = enable_typechecks
         self._parent_join_cache = {}
+        self.comparator = PropertyLoader.Comparator(self)
+        self.join_depth = join_depth
+        self.strategy_class = strategy_class
 
         if cascade is not None:
             self.cascade = mapperutil.CascadeOptions(cascade)
         else:
             if private:
+                util.warn_deprecated('private option is deprecated; see docs for details')
                 self.cascade = mapperutil.CascadeOptions("all, delete-orphan")
             else:
                 self.cascade = mapperutil.CascadeOptions("save-update, merge")
+        
+        if self.passive_deletes == 'all' and ("delete" in self.cascade or "delete-orphan" in self.cascade):
+            raise exceptions.ArgumentError("Can't set passive_deletes='all' in conjunction with 'delete' or 'delete-orphan' cascade")
 
         self.association = association
+        if association:
+            util.warn_deprecated('association option is deprecated; see docs for details')
         self.order_by = order_by
         self.attributeext=attributeext
         if isinstance(backref, str):
@@ -104,13 +182,95 @@ class PropertyLoader(StrategizedProperty):
             self.backref = backref
         self.is_backref = is_backref
 
-    def compare(self, value):
-        return sql.and_(*[x==y for (x, y) in zip(self.mapper.primary_key, self.mapper.primary_key_from_instance(value))])
+    class Comparator(PropComparator):
+        def __eq__(self, other):
+            if other is None:
+                return ~sql.exists([1], self.prop.primaryjoin)
+            elif self.prop.uselist:
+                if not hasattr(other, '__iter__'):
+                    raise exceptions.InvalidRequestError("Can only compare a collection to an iterable object.  Use contains().")
+                else:
+                    j = self.prop.primaryjoin
+                    if self.prop.secondaryjoin:
+                        j = j & self.prop.secondaryjoin
+                    clauses = []
+                    for o in other:
+                        clauses.append(
+                            sql.exists([1], j & sql.and_(*[x==y for (x, y) in zip(self.prop.mapper.primary_key, self.prop.mapper.primary_key_from_instance(o))]))
+                        )
+                    return sql.and_(*clauses)
+            else:  
+                return self.prop._optimized_compare(other)
+        
+        def any(self, criterion=None, **kwargs):
+            if not self.prop.uselist:
+                raise exceptions.InvalidRequestError("'any()' not implemented for scalar attributes. Use has().")
+            j = self.prop.primaryjoin
+            if self.prop.secondaryjoin:
+                j = j & self.prop.secondaryjoin
+            for k in kwargs:
+                crit = (getattr(self.prop.mapper.class_, k) == kwargs[k])
+                if criterion is None:
+                    criterion = crit
+                else:
+                    criterion = criterion & crit
+            return sql.exists([1], j & criterion)
+        
+        def has(self, criterion=None, **kwargs):
+            if self.prop.uselist:
+                raise exceptions.InvalidRequestError("'has()' not implemented for collections.  Use any().")
+            j = self.prop.primaryjoin
+            if self.prop.secondaryjoin:
+                j = j & self.prop.secondaryjoin
+            for k in kwargs:
+                crit = (getattr(self.prop.mapper.class_, k) == kwargs[k])
+                if criterion is None:
+                    criterion = crit
+                else:
+                    criterion = criterion & crit
+            return sql.exists([1], j & criterion)
+                
+        def contains(self, other):
+            if not self.prop.uselist:
+                raise exceptions.InvalidRequestError("'contains' not implemented for scalar attributes.  Use ==")
+            clause = self.prop._optimized_compare(other)
 
+            if self.prop.secondaryjoin:
+                j = self.prop.primaryjoin
+                j = j & self.prop.secondaryjoin
+                clause.negation_clause = ~sql.exists([1], j & sql.and_(*[x==y for (x, y) in zip(self.prop.mapper.primary_key, self.prop.mapper.primary_key_from_instance(other))]))
+
+            return clause
+
+        def __ne__(self, other):
+            if self.prop.uselist and not hasattr(other, '__iter__'):
+                raise exceptions.InvalidRequestError("Can only compare a collection to an iterable object")
+                
+            j = self.prop.primaryjoin
+            if self.prop.secondaryjoin:
+                j = j & self.prop.secondaryjoin
+            return ~sql.exists([1], j & sql.and_(*[x==y for (x, y) in zip(self.prop.mapper.primary_key, self.prop.mapper.primary_key_from_instance(other))]))
+            
+    def compare(self, op, value, value_is_parent=False):
+        if op == operator.eq:
+            if value is None:
+                return ~sql.exists([1], self.prop.mapper.mapped_table, self.prop.primaryjoin)
+            else:
+                return self._optimized_compare(value, value_is_parent=value_is_parent)
+        else:
+            return op(self.comparator, value)
+    
+    def _optimized_compare(self, value, value_is_parent=False):
+        return self._get_strategy(strategies.LazyLoader).lazy_clause(value, reverse_direction=not value_is_parent)
+        
     private = property(lambda s:s.cascade.delete_orphan)
 
     def create_strategy(self):
-        if self.lazy:
+        if self.strategy_class:
+            return self.strategy_class(self)
+        elif self.lazy == 'dynamic':
+            return strategies.DynaLoader(self)
+        elif self.lazy:
             return strategies.LazyLoader(self)
         elif self.lazy is False:
             return strategies.EagerLoader(self)
@@ -120,26 +280,27 @@ class PropertyLoader(StrategizedProperty):
     def __str__(self):
         return str(self.parent.class_.__name__) + "." + self.key + " (" + str(self.mapper.class_.__name__)  + ")"
 
-    def merge(self, session, source, dest, _recursive):
+    def merge(self, session, source, dest, dont_load, _recursive):
         if not "merge" in self.cascade or self.mapper in _recursive:
             return
         childlist = sessionlib.attribute_manager.get_history(source, self.key, passive=True)
         if childlist is None:
             return
         if self.uselist:
-            # sets a blank list according to the correct list class
-            dest_list = getattr(self.parent.class_, self.key).initialize(dest)
+            # sets a blank collection according to the correct list class
+            dest_list = sessionlib.attribute_manager.init_collection(dest, self.key)
             for current in list(childlist):
-                obj = session.merge(current, entity_name=self.mapper.entity_name, _recursive=_recursive)
+                obj = session.merge(current, entity_name=self.mapper.entity_name, dont_load=dont_load, _recursive=_recursive)
                 if obj is not None:
-                    dest_list.append(obj)
+                    dest_list.append_with_event(obj)
         else:
             current = list(childlist)[0]
             if current is not None:
-                obj = session.merge(current, entity_name=self.mapper.entity_name, _recursive=_recursive)
+                obj = session.merge(current, entity_name=self.mapper.entity_name, dont_load=dont_load, _recursive=_recursive)
                 if obj is not None:
                     setattr(dest, self.key, obj)
 
+
     def cascade_iterator(self, type, object, recursive, halt_on=None):
         if not type in self.cascade:
             return
@@ -252,7 +413,7 @@ class PropertyLoader(StrategizedProperty):
                 raise exceptions.ArgumentError("In relationship '%s', primary and secondary join conditions must not include columns from the polymorphic 'select_table' argument as of SA release 0.3.4.  Construct join conditions using the base tables of the related mappers." % (str(self)))
 
     def _determine_fks(self):
-        if len(self._legacy_foreignkey) and not self._is_self_referential():
+        if self._legacy_foreignkey and not self._is_self_referential():
             self.foreign_keys = self._legacy_foreignkey
 
         def col_is_part_of_mappings(col):
@@ -264,10 +425,10 @@ class PropertyLoader(StrategizedProperty):
                     self.target.corresponding_column(col, raiseerr=False) is not None or \
                     self.secondary.corresponding_column(col, raiseerr=False) is not None
 
-        if len(self.foreign_keys):
+        if self.foreign_keys:
             self._opposite_side = util.Set()
             def visit_binary(binary):
-                if binary.operator != '=' or not isinstance(binary.left, schema.Column) or not isinstance(binary.right, schema.Column):
+                if binary.operator != operator.eq or not isinstance(binary.left, schema.Column) or not isinstance(binary.right, schema.Column):
                     return
                 if binary.left in self.foreign_keys:
                     self._opposite_side.add(binary.right)
@@ -280,7 +441,7 @@ class PropertyLoader(StrategizedProperty):
             self.foreign_keys = util.Set()
             self._opposite_side = util.Set()
             def visit_binary(binary):
-                if binary.operator != '=' or not isinstance(binary.left, schema.Column) or not isinstance(binary.right, schema.Column):
+                if binary.operator != operator.eq or not isinstance(binary.left, schema.Column) or not isinstance(binary.right, schema.Column):
                     return
 
                 # this check is for when the user put the "view_only" flag on and has tables that have nothing
@@ -320,14 +481,14 @@ class PropertyLoader(StrategizedProperty):
             # for a self referential mapper, if the "foreignkey" is a single or composite primary key,
             # then we are "many to one", since the remote site of the relationship identifies a singular entity.
             # otherwise we are "one to many".
-            if len(self._legacy_foreignkey):
+            if self._legacy_foreignkey:
                 for f in self._legacy_foreignkey:
                     if not f.primary_key:
                         self.direction = sync.ONETOMANY
                     else:
                         self.direction = sync.MANYTOONE
 
-            elif len(self.remote_side):
+            elif self.remote_side:
                 for f in self.foreign_keys:
                     if f in self.remote_side:
                         self.direction = sync.ONETOMANY
@@ -362,16 +523,13 @@ class PropertyLoader(StrategizedProperty):
                     "argument." % (str(self)))
 
     def _determine_remote_side(self):
-        if len(self.remote_side):
-            return
-        self.remote_side = util.Set()
+        if not self.remote_side:
+            if self.direction is sync.MANYTOONE:
+                self.remote_side = util.Set(self._opposite_side)
+            elif self.direction is sync.ONETOMANY or self.direction is sync.MANYTOMANY:
+                self.remote_side = util.Set(self.foreign_keys)
 
-        if self.direction is sync.MANYTOONE:
-            for c in self._opposite_side:
-                self.remote_side.add(c)
-        elif self.direction is sync.ONETOMANY or self.direction is sync.MANYTOMANY:
-            for c in self.foreign_keys:
-                self.remote_side.add(c)
+        self.local_side = util.Set(self._opposite_side).union(util.Set(self.foreign_keys)).difference(self.remote_side)
 
     def _create_polymorphic_joins(self):
         # get ready to create "polymorphic" primary/secondary join clauses.
@@ -383,27 +541,26 @@ class PropertyLoader(StrategizedProperty):
         # as we will be using the polymorphic selectables (i.e. select_table argument to Mapper) to figure this out,
         # first create maps of all the "equivalent" columns, since polymorphic selectables will often munge
         # several "equivalent" columns (such as parent/child fk cols) into just one column.
-        target_equivalents = self.mapper._get_inherited_column_equivalents()
 
+        target_equivalents = self.mapper._get_equivalent_columns()
+            
         # if the target mapper loads polymorphically, adapt the clauses to the target's selectable
         if self.loads_polymorphic:
             if self.secondaryjoin:
-                self.polymorphic_secondaryjoin = self.secondaryjoin.copy_container()
-                sql_util.ClauseAdapter(self.mapper.select_table).traverse(self.polymorphic_secondaryjoin)
-                self.polymorphic_primaryjoin = self.primaryjoin.copy_container()
+                self.polymorphic_secondaryjoin = sql_util.ClauseAdapter(self.mapper.select_table).traverse(self.secondaryjoin, clone=True)
+                self.polymorphic_primaryjoin = self.primaryjoin
             else:
-                self.polymorphic_primaryjoin = self.primaryjoin.copy_container()
                 if self.direction is sync.ONETOMANY:
-                    sql_util.ClauseAdapter(self.mapper.select_table, include=self.foreign_keys, equivalents=target_equivalents).traverse(self.polymorphic_primaryjoin)
+                    self.polymorphic_primaryjoin = sql_util.ClauseAdapter(self.mapper.select_table, include=self.foreign_keys, equivalents=target_equivalents).traverse(self.primaryjoin, clone=True)
                 elif self.direction is sync.MANYTOONE:
-                    sql_util.ClauseAdapter(self.mapper.select_table, exclude=self.foreign_keys, equivalents=target_equivalents).traverse(self.polymorphic_primaryjoin)
+                    self.polymorphic_primaryjoin = sql_util.ClauseAdapter(self.mapper.select_table, exclude=self.foreign_keys, equivalents=target_equivalents).traverse(self.primaryjoin, clone=True)
                 self.polymorphic_secondaryjoin = None
             # load "polymorphic" versions of the columns present in "remote_side" - this is
             # important for lazy-clause generation which goes off the polymorphic target selectable
             for c in list(self.remote_side):
-                if self.secondary and c in self.secondary.columns:
+                if self.secondary and self.secondary.columns.contains_column(c):
                     continue
-                for equiv in [c] + (c in target_equivalents and target_equivalents[c] or []): 
+                for equiv in [c] + (c in target_equivalents and list(target_equivalents[c]) or []): 
                     corr = self.mapper.select_table.corresponding_column(equiv, raiseerr=False)
                     if corr:
                         self.remote_side.add(corr)
@@ -411,13 +568,15 @@ class PropertyLoader(StrategizedProperty):
                 else:
                     raise exceptions.AssertionError(str(self) + ": Could not find corresponding column for " + str(c) + " in selectable "  + str(self.mapper.select_table))
         else:
-            self.polymorphic_primaryjoin = self.primaryjoin.copy_container()
-            self.polymorphic_secondaryjoin = self.secondaryjoin and self.secondaryjoin.copy_container() or None
+            self.polymorphic_primaryjoin = self.primaryjoin
+            self.polymorphic_secondaryjoin = self.secondaryjoin
 
     def _post_init(self):
         if logging.is_info_enabled(self.logger):
             self.logger.info(str(self) + " setup primary join " + str(self.primaryjoin))
             self.logger.info(str(self) + " setup polymorphic primary join " + str(self.polymorphic_primaryjoin))
+            self.logger.info(str(self) + " setup secondary join " + str(self.secondaryjoin))
+            self.logger.info(str(self) + " setup polymorphic secondary join " + str(self.polymorphic_secondaryjoin))
             self.logger.info(str(self) + " foreign keys " + str([str(c) for c in self.foreign_keys]))
             self.logger.info(str(self) + " remote columns " + str([str(c) for c in self.remote_side]))
             self.logger.info(str(self) + " relation direction " + (self.direction is sync.ONETOMANY and "one-to-many" or (self.direction is sync.MANYTOONE and "many-to-one" or "many-to-many")))
@@ -440,7 +599,7 @@ class PropertyLoader(StrategizedProperty):
 
             if self.backref is not None:
                 self.backref.compile(self)
-        elif not sessionlib.attribute_manager.is_class_managed(self.parent.class_, self.key):
+        elif not mapper.class_mapper(self.parent.class_).get_property(self.key, raiseerr=False):
             raise exceptions.ArgumentError("Attempting to assign a new relation '%s' to a non-primary mapper on class '%s'.  New relations can only be added to the primary mapper, i.e. the very first mapper created for class '%s' " % (self.key, self.parent.class_.__name__, self.parent.class_.__name__))
 
         super(PropertyLoader, self).do_init()
@@ -448,22 +607,39 @@ class PropertyLoader(StrategizedProperty):
     def _is_self_referential(self):
         return self.parent.mapped_table is self.target or self.parent.select_table is self.target
 
-    def get_join(self, parent, primary=True, secondary=True):
+    def get_join(self, parent, primary=True, secondary=True, polymorphic_parent=True):
+        """return a join condition from the given parent mapper to this PropertyLoader's mapper.
+        
+           The resulting ClauseElement object is cached and should not be modified directly.
+        
+            parent
+              a mapper which has a relation() to this PropertyLoader.  A PropertyLoader can 
+              have multiple "parents" when its actual parent mapper has inheriting mappers.
+              
+            primary
+              include the primary join condition in the resulting join.
+              
+            secondary
+              include the secondary join condition in the resulting join.  If both primary
+              and secondary are returned, they are joined via AND.
+              
+            polymorphic_parent
+              if True, use the parent's 'select_table' instead of its 'mapped_table' to produce the join.
+        """
+        
         try:
-            return self._parent_join_cache[(parent, primary, secondary)]
+            return self._parent_join_cache[(parent, primary, secondary, polymorphic_parent)]
         except KeyError:
-            parent_equivalents = parent._get_inherited_column_equivalents()
-            primaryjoin = self.polymorphic_primaryjoin.copy_container()
-            if self.secondaryjoin is not None:
-                secondaryjoin = self.polymorphic_secondaryjoin.copy_container()
-            else:
-                secondaryjoin = None
-            if self.direction is sync.ONETOMANY:
-                sql_util.ClauseAdapter(parent.select_table, exclude=self.foreign_keys, equivalents=parent_equivalents).traverse(primaryjoin)
-            elif self.direction is sync.MANYTOONE:
-                sql_util.ClauseAdapter(parent.select_table, include=self.foreign_keys, equivalents=parent_equivalents).traverse(primaryjoin)
-            elif self.secondaryjoin:
-                sql_util.ClauseAdapter(parent.select_table, exclude=self.foreign_keys, equivalents=parent_equivalents).traverse(primaryjoin)
+            parent_equivalents = parent._get_equivalent_columns()
+            secondaryjoin = self.polymorphic_secondaryjoin
+            if polymorphic_parent:
+                # adapt the "parent" side of our join condition to the "polymorphic" select of the parent
+                if self.direction is sync.ONETOMANY:
+                    primaryjoin = sql_util.ClauseAdapter(parent.select_table, exclude=self.foreign_keys, equivalents=parent_equivalents).traverse(self.polymorphic_primaryjoin, clone=True)
+                elif self.direction is sync.MANYTOONE:
+                    primaryjoin = sql_util.ClauseAdapter(parent.select_table, include=self.foreign_keys, equivalents=parent_equivalents).traverse(self.polymorphic_primaryjoin, clone=True)
+                elif self.secondaryjoin:
+                    primaryjoin = sql_util.ClauseAdapter(parent.select_table, exclude=self.foreign_keys, equivalents=parent_equivalents).traverse(self.polymorphic_primaryjoin, clone=True)
 
             if secondaryjoin is not None:
                 if secondary and not primary:
@@ -474,7 +650,7 @@ class PropertyLoader(StrategizedProperty):
                     j = primaryjoin
             else:
                 j = primaryjoin
-            self._parent_join_cache[(parent, primary, secondary)] = j
+            self._parent_join_cache[(parent, primary, secondary, polymorphic_parent)] = j
             return j
 
     def register_dependencies(self, uowcommit):
@@ -484,54 +660,47 @@ class PropertyLoader(StrategizedProperty):
 PropertyLoader.logger = logging.class_logger(PropertyLoader)
 
 class BackRef(object):
-    """Stores the name of a backreference property as well as options
-    to be used on the resulting PropertyLoader.
-    """
+    """Attached to a PropertyLoader to indicate a complementary reverse relationship.
+    
+    Can optionally create the complementing PropertyLoader if one does not exist already."""
 
-    def __init__(self, key, **kwargs):
+    def __init__(self, key, _prop=None, **kwargs):
         self.key = key
         self.kwargs = kwargs
+        self.prop = _prop
 
     def compile(self, prop):
-        """Called by the owning PropertyLoader to set up a
-        backreference on the PropertyLoader's mapper.
-        """
-
-        # try to set a LazyLoader on our mapper referencing the parent mapper
+        if self.prop:
+            return
+        
+        self.prop = prop
+        
         mapper = prop.mapper.primary_mapper()
-        if not mapper.props.has_key(self.key):
+        if mapper.get_property(self.key, raiseerr=False) is None:
             pj = self.kwargs.pop('primaryjoin', None)
             sj = self.kwargs.pop('secondaryjoin', None)
-            # the backref property is set on the primary mapper
+
             parent = prop.parent.primary_mapper()
             self.kwargs.setdefault('viewonly', prop.viewonly)
             self.kwargs.setdefault('post_update', prop.post_update)
+                
             relation = PropertyLoader(parent, prop.secondary, pj, sj,
-                                      backref=prop.key, is_backref=True,
+                                      backref=BackRef(prop.key, _prop=prop), 
+                                      is_backref=True,
                                       **self.kwargs)
+                                      
             mapper._compile_property(self.key, relation);
-        elif not isinstance(mapper.props[self.key], PropertyLoader):
-            raise exceptions.ArgumentError(
-                "Can't create backref '%s' on mapper '%s'; an incompatible "
-                "property of that name already exists" % (self.key, str(mapper)))
+
+            prop.reverse_property = mapper.get_property(self.key)
+            mapper.get_property(self.key).reverse_property = prop
+
         else:
-            # else set one of us as the "backreference"
-            parent = prop.parent.primary_mapper()
-            if parent.class_ is not mapper.props[self.key]._get_target_class():
-                raise exceptions.ArgumentError(
-                    "Backrefs do not match:  backref '%s' expects to connect to %s, "
-                    "but found a backref already connected to %s" %
-                    (self.key, str(parent.class_), str(mapper.props[self.key].mapper.class_)))
-            if not mapper.props[self.key].is_backref:
-                prop.is_backref=True
-                if not prop.viewonly:
-                    prop._dependency_processor.is_backref=True
-                    # reverse_property used by dependencies.ManyToManyDP to check
-                    # association table operations
-                    prop.reverse_property = mapper.props[self.key]
-                    mapper.props[self.key].reverse_property = prop
+            raise exceptions.ArgumentError("Error creating backref '%s' on relation '%s': property of that name exists on mapper '%s'" % (self.key, prop, mapper))
 
     def get_extension(self):
         """Return an attribute extension to use with this backreference."""
 
         return attributes.GenericBackrefExtension(self.key)
+
+mapper.ColumnProperty = ColumnProperty
+        
